// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.fsm;

import tango.core.Array;
import tango.io.Stdout;
import tango.text.Util;
import Integer = tango.text.convert.Integer;

import semitwist.util;

import stc.util;

enum RegexOpMultiplicity
{
	One,
	Any
}

enum RegexOpInclusivity
{
	Include,
	Exclude
}
	
class RegexOp(T)
{
	RegexOpMultiplicity multiplicity;
	RegexOpInclusivity inclusivity;
	T[] inputs;
	
	this(RegexOpMultiplicity multiplicity, RegexOpInclusivity inclusivity, T[] inputs)
	{
		// Bugzilla #1610
		// http://d.puremagic.com/issues/show_bug.cgi?id=1610
		//mixin(initMember!(multiplicity, inclusivity, inputs));
		this.multiplicity = multiplicity;
		this.inclusivity = inclusivity;
		mixin(initMember!(inputs));
		
		// Required by tango.core.Array.intersectionOf() and (same).missingFrom(),
		// both used in this.intersectionWith().
		sort(this.inputs);
		
		// Don't need duplicates, they'll just slow down parsing
//Stdout.formatln("------");
//mixin(traceVal!("inputs"));
		   // "unique" is renamed "distinct" in Tango 0.99.7
		//this.inputs = this.inputs[0..distinct(this.inputs)];
		this.inputs = this.inputs[0..distinct(this.inputs)];
//mixin(traceVal!("inputs"));
	}
	
	static RegexOpInclusivity invert(RegexOpInclusivity incl)
	{
		return incl == RegexOpInclusivity.Include ?
		       RegexOpInclusivity.Exclude :
			   RegexOpInclusivity.Include;
	}
	
	RegexOpInclusivity reverseOfInclusivity()
	{
		return invert(inclusivity);
	}
	
	RegexOp!(T) invert()
	{
		return new RegexOp!(T)(multiplicity, invert(inclusivity), inputs);
	}
	
	RegexOp!(T) intersect(RegexOp!(T) op)
	{
		T[] retInputs;
		RegexOpInclusivity retInclusivity;
		
		if(this.inclusivity == RegexOpInclusivity.Include &&
		     op.inclusivity == RegexOpInclusivity.Include )
		{
			retInputs = intersectionOf(this.inputs, op.inputs);
			retInclusivity = RegexOpInclusivity.Include;
		}
		else if(this.inclusivity == RegexOpInclusivity.Exclude &&
		          op.inclusivity == RegexOpInclusivity.Exclude )
		{
			retInputs = this.inputs ~ op.inputs;
			retInclusivity = RegexOpInclusivity.Exclude;
		}
		else if(this.inclusivity == RegexOpInclusivity.Include &&
		          op.inclusivity == RegexOpInclusivity.Exclude )
		{
			retInputs = missingFrom(this.inputs, op.inputs);
			retInclusivity = RegexOpInclusivity.Include;
		}
		else
		{
			retInputs = missingFrom(op.inputs, this.inputs);
			retInclusivity = RegexOpInclusivity.Include;
		}
		
		return new RegexOp!(T)(RegexOpMultiplicity.One, retInclusivity, retInputs);
	}
	
	RegexOp!(T) minus(RegexOp!(T) op)
	{
		T[] retInputs;
		RegexOpInclusivity retInclusivity;
		
		if(this.inclusivity == RegexOpInclusivity.Include &&
		     op.inclusivity == RegexOpInclusivity.Include )
		{
			retInputs = missingFrom(this.inputs, op.inputs);
			retInclusivity = RegexOpInclusivity.Include;
		}
		else if(this.inclusivity == RegexOpInclusivity.Exclude &&
		          op.inclusivity == RegexOpInclusivity.Exclude )
		{
			retInputs = missingFrom(op.inputs, this.inputs);
			retInclusivity = RegexOpInclusivity.Include;
		}
		else if(this.inclusivity == RegexOpInclusivity.Include &&
		          op.inclusivity == RegexOpInclusivity.Exclude )
		{
			retInputs = intersectionOf(this.inputs, op.inputs);
			retInclusivity = RegexOpInclusivity.Include;
		}
		else
		{
			retInputs = this.inputs ~ op.inputs;
			retInclusivity = RegexOpInclusivity.Exclude;
		}
		
		return new RegexOp!(T)(RegexOpMultiplicity.One, retInclusivity, retInputs);
	}

	RegexOp!(T) plus(RegexOp!(T) op)
	{
		//throw new Exception("FSM(T).plus(RegexOp!(T)) Not yet implemented");
		return op.invert().minus(this).invert();
	}
	
	unittest
	{
		RegexOp!(char) opA;
		RegexOp!(char) opB;
		RegexOp!(char) opASave;
		RegexOp!(char) opBSave;
		RegexOp!(char) opResult;
		char[] sortedResultInputs;
		
		void EnsureNoChanges(size_t lineNum)
		{
			NonFatalAssert!(__LINE__,__FILE__)(opA.multiplicity == opASave.multiplicity, Integer.toString(lineNum));
			NonFatalAssert!(__LINE__,__FILE__)(opA.inclusivity  == opASave.inclusivity,  Integer.toString(lineNum));
			NonFatalAssert!(__LINE__,__FILE__)(opA.inputs       == opASave.inputs,       Integer.toString(lineNum));
			
			NonFatalAssert!(__LINE__,__FILE__)(opB.multiplicity == opBSave.multiplicity, Integer.toString(lineNum));
			NonFatalAssert!(__LINE__,__FILE__)(opB.inclusivity  == opBSave.inclusivity,  Integer.toString(lineNum));
			NonFatalAssert!(__LINE__,__FILE__)(opB.inputs       == opBSave.inputs,       Integer.toString(lineNum));
		}
		
		// Test invert()
		opA     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "AB");
		opASave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "AB");
		opB     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "BC");
		opBSave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "BC");
		opResult = opA.invert();
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "AB", `"`~opResult.inputs~`"`);
		opResult = opB.invert();
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "BC", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);


		// Both RegexOpInclusivity.Include
		opA     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "AB");
		opASave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "AB");
		opB     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "BC");
		opBSave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "BC");
		opResult = opA.intersect(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "B", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.intersect(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "B", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.minus(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "A", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.minus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "C", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.minus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.plus(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "ABBC" || sortedResultInputs == "ABC", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.plus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "ABBC" || sortedResultInputs == "ABC", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.plus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "AABB" || sortedResultInputs == "AB", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		// Both RegexOpInclusivity.Exclude
		opA     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "AB");
		opASave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "AB");
		opB     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "BC");
		opBSave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "BC");
		opResult = opA.intersect(opB);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "ABC" || sortedResultInputs == "ABBC", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.intersect(opA);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "ABC" || sortedResultInputs == "ABBC", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.minus(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "C", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.minus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "A", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.minus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.plus(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "B", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.plus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "B", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.plus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "AABB" || sortedResultInputs == "AB", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		// A = RegexOpInclusivity.Include, B = RegexOpInclusivity.Exclude
		opA     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "AB");
		opASave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Include, "AB");
		opB     = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "BC");
		opBSave = new RegexOp!(char)(RegexOpMultiplicity.One, RegexOpInclusivity.Exclude, "BC");
		opResult = opA.intersect(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "A", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.intersect(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "A", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.minus(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Include);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "B", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.minus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		sortedResultInputs = opResult.inputs;
		sort(sortedResultInputs);
		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "ABC", `"`~sortedResultInputs~`"`);
//		NonFatalAssert!(__LINE__,__FILE__)(sortedResultInputs == "ABC" || sortedResultInputs == "ABBC", `"`~sortedResultInputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opA.plus(opB);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "C", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
		opResult = opB.plus(opA);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.multiplicity == RegexOpMultiplicity.One);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inclusivity == RegexOpInclusivity.Exclude);
		NonFatalAssert!(__LINE__,__FILE__)(opResult.inputs == "C", `"`~opResult.inputs~`"`);
		EnsureNoChanges(__LINE__);
		
	}
}

class FSM(T)
{
	private FSMState!(T) initState;
	private FSMState!(T) currState;
	
	this(RegexOp!(T)[] regexOps)
	{
//		throw new Exception("FSM(T).this(RegexOp!(T)[]) Not yet implemented");
	
		const char[] unknownMultMsg = "FSM(T).this(RegexOp!(T)[]): Unknown RegexOpMultiplicity";
		FSMState!(T)[] states;
		RegexOp!(T)[][] chart;
		RegexOp!(T) prevOp;
		RegexOp!(T) prevOneOp = null;
		
		// Inclusive
		size_t chartEndFrom;
		size_t chartEndTo;

		switch(regexOps[$-1].multiplicity)
		{
		case RegexOpMultiplicity.One:
			//Fill out chart with regexOps[$-1], ver A
			chartEndFrom = 1;
			chartEndTo   = 2;
			chart.length = chartEndFrom+1;
			for(size_t fromIndex; fromIndex <= chartEndFrom; fromIndex++)
				chart[fromIndex].length = chartEndTo+1;
			chart[0][0] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                              RegexOpInclusivity.Exclude,
										  cast(T[])[]);
			chart[1][1] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                              regexOps[$-1].inclusivity,
										  regexOps[$-1].inputs);
			prevOneOp = regexOps[$-1];
			break;
		
		case RegexOpMultiplicity.Any:
			//Fill out chart with regexOps[$-1], ver [A]*
			chartEndFrom = 0;
			chartEndTo   = 1;
			chart.length = chartEndFrom+1;
			for(size_t fromIndex; fromIndex <= chartEndFrom; fromIndex++)
				chart[fromIndex].length = chartEndTo+1;
			chart[0][0] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                              regexOps[$-1].reverseOfInclusivity(),
										  regexOps[$-1].inputs);

/*			chart[1][0] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                              regexOps[$-1].inclusivity,
										  regexOps[$-1].inputs);
*/
			chart[0][1] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                              regexOps[$-1].inclusivity,
										  regexOps[$-1].inputs);

			break;
		
		default:
			throw new Exception(unknownMultMsg);
		}
		prevOp = regexOps[$-1];

		foreach_reverse(RegexOp!(T) op; regexOps[0..$-1])
		{
			if(op.multiplicity == RegexOpMultiplicity.One)
			{
				//Fill out chart (From notes, line 2)
				chartEndFrom++;
				chartEndTo++;
				chart.length = chartEndFrom+1;
				for(size_t fromIndex; fromIndex <= chartEndFrom; fromIndex++)
					chart[fromIndex].length = chartEndTo+1;
/*				chart[chartEndTo-1][chartEndFrom] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                                                        op.inclusivity,
				                                                    op.inputs);
*/
				chart[chartEndFrom][chartEndTo-1] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                                                        op.inclusivity,
				                                                    op.inputs);

				prevOneOp = op;
			}
			else if(op.multiplicity == RegexOpMultiplicity.Any)
			{
				if(prevOp.multiplicity == RegexOpMultiplicity.One)
				{
					//Fill out chart (From notes, line 5)
					chartEndFrom++;
					chartEndTo++;
					chart.length = chartEndFrom+1;
					for(size_t fromIndex; fromIndex <= chartEndFrom; fromIndex++)
						chart[fromIndex].length = chartEndTo+1;
/*					chart[0][chartEndFrom-1] = op.plus(prevOp).invert();
					chart[chartEndTo-2][chartEndFrom-1] = chart[chartEndFrom-1][chartEndTo-2].minus(op);
					chart[chartEndTo-1][chartEndFrom-1] = op.intersect(prevOp);
					chart[chartEndTo-0][chartEndFrom-1] = op.minus(prevOp);

					chart[chartEndTo-2][chartEndFrom-0] = chart[chartEndFrom-1][chartEndTo-2];
					chart[chartEndTo-1][chartEndFrom-0] = chart[chartEndFrom-1][chartEndTo-1];
					chart[chartEndTo-0][chartEndFrom-0] = chart[chartEndFrom-1][chartEndTo-0];
*/
					chart[chartEndFrom-1][chartEndTo-2] = op.plus(prevOp).invert();
//					chart[chartEndFrom-1][chartEndTo-2] = chart[chartEndFrom-1][chartEndTo-2].minus(op);
					chart[chartEndFrom-1][chartEndTo-2] = prevOp.minus(op);
					chart[chartEndFrom-1][chartEndTo-1] = op.intersect(prevOp);
					chart[chartEndFrom-1][chartEndTo-0] = op.minus(prevOp);

					chart[chartEndFrom-0][chartEndTo-2] = chart[chartEndFrom-1][chartEndTo-2];
					chart[chartEndFrom-0][chartEndTo-1] = chart[chartEndFrom-1][chartEndTo-1];
					chart[chartEndFrom-0][chartEndTo-0] = chart[chartEndFrom-1][chartEndTo-0];

				}
				else if(prevOp.multiplicity == RegexOpMultiplicity.Any)
				{
					if(prevOneOp is null)
					{
						//Fill out chart (From notes, line 3)
						chartEndFrom++;
						chartEndTo++;
						chart.length = chartEndFrom+1;
						for(size_t fromIndex; fromIndex <= chartEndFrom; fromIndex++)
							chart[fromIndex].length = chartEndTo+1;
/*						for(size_t i = 0; i < chartEndTo-1; i++)
							chart[chartEndFrom-0][i] = chart[chartEndFrom-1][i].minus(op);
						chart[chartEndFrom-0][chartEndTo-0] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                                                                  op.inclusivity,
				                                                              op.inputs);
*/
						for(size_t i = 0; i < chartEndTo-1; i++)
							chart[chartEndFrom-0][i] = chart[chartEndFrom-1][i].minus(op);
						chart[chartEndFrom-0][chartEndTo-0] = new RegexOp!(T)(RegexOpMultiplicity.One,
			                                                                  op.inclusivity,
				                                                              op.inputs);

					}
					else
					{
						//Fill out chart (From notes, line 4)
/*						chart[chartEndFrom-0][chartEndTo-2] = chart[chartEndFrom-0][chartEndTo-2].minus(op);
						chart[chartEndFrom-0][chartEndTo-1] =
								chart[chartEndFrom-0][chartEndTo-1].plus(op.intersect(prevOneOp));
						chart[chartEndFrom-0][chartEndTo-0] = 
								chart[chartEndFrom-0][chartEndTo-1].plus(op.minus(prevOneOp));
								
						chart[chartEndFrom-1][chartEndTo-2] = chart[chartEndFrom-0][chartEndTo-2];
						chart[chartEndFrom-1][chartEndTo-1] = chart[chartEndFrom-0][chartEndTo-1];
						chart[chartEndFrom-1][chartEndTo-0] = chart[chartEndFrom-0][chartEndTo-0];
						
						chart[chartEndFrom-1][0] = chart[chartEndFrom-1][0].plus(op);
*/

						chart[chartEndFrom-0][chartEndTo-2] = chart[chartEndFrom-0][chartEndTo-2].minus(op);
						chart[chartEndFrom-0][chartEndTo-1] =
								chart[chartEndFrom-0][chartEndTo-1].plus(op.intersect(prevOneOp));
						chart[chartEndFrom-0][chartEndTo-0] = 
								chart[chartEndFrom-0][chartEndTo-1].plus(op.minus(prevOneOp));
								
						chart[chartEndFrom-1][chartEndTo-2] = chart[chartEndFrom-0][chartEndTo-2];
						chart[chartEndFrom-1][chartEndTo-1] = chart[chartEndFrom-0][chartEndTo-1];
						chart[chartEndFrom-1][chartEndTo-0] = chart[chartEndFrom-0][chartEndTo-0];
						
						chart[chartEndFrom-1][0] = chart[chartEndFrom-1][0].plus(op);

					}
				}
				else
					throw new Exception(unknownMultMsg);
			}
			else
				throw new Exception(unknownMultMsg);
				
			
			prevOp = op;
			
		//Convert chart to FSMStates and FSM
		states.length = chartEndFrom+1;
		
		for(size_t i=0; i < chartEndFrom+1; i++)
		{
			states[i] = new FSMState!(T)();
			states[i].thisStatesIndex = i;
		}
			
		for(size_t indexFrom=0; indexFrom <= chartEndFrom; indexFrom++)
		{
		foreach(size_t indexTo, RegexOp!(T) _op; chart[indexFrom])
		{
			auto currOp = chart[chartEndFrom-indexFrom][chartEndTo-indexTo];
			auto destState = indexTo == chartEndTo ? FSMState!(T).endState : states[indexTo];
			if(currOp !is null)
				states[indexFrom].addNextState(currOp, destState);
		}
		}
		output(states);
		}
		
		//Convert chart to FSMStates and FSM
		states.length = chartEndFrom+1;
		
		for(size_t i=0; i < chartEndFrom+1; i++)
		{
			states[i] = new FSMState!(T)();
			states[i].thisStatesIndex = i;
		}
			
		for(size_t indexFrom=0; indexFrom <= chartEndFrom; indexFrom++)
		foreach(size_t indexTo, RegexOp!(T) _op; chart[indexFrom])
		{
			auto currOp = chart[chartEndFrom-indexFrom][chartEndTo-indexTo];
			auto destState = indexTo == chartEndTo ? FSMState!(T).endState : states[indexTo];
			if(currOp !is null)
				states[indexFrom].addNextState(currOp, destState);
		}
//Stdout.formatln("blah: {} (line {})", __FILE__, __LINE__);
		output(states);
//Stdout.formatln("blah: {} (line {})", __FILE__, __LINE__);
			
		initState = states[0];
		currState = states[0];
		
	}
	
/*	this(RegexOp!(T) opA, RegexOp!(T) opB)
	{
		throw new Exception("FSM(T).this(RegexOp!(T), RegexOp!(T)) Not yet implemented");

		size_t numStates;
		FSMState!(schar)[] fsm;
		
		switch(opA.multiplicity)
		{
		case RegexOpMultiplicity.One:
			for(size_t i=0; i < 2; i++)
				fsm ~= new FSMState!(T)();
			
			fsm[0].addNextState(opA.inclusivity, opA.inputs, fsm[1]);

			fsm[1].addNextState(RegexOpInclusivity.Include, opA.inputs, FSMState!(T).endState);
			fsm[1].addNextState(RegexOpInclusivity.Exclude, opA.inputs, FSMState!(T).endState);
			break;
			
		case RegexOpMultiplicity.Any:
			for(size_t i=0; i < 2; i++)
				fsm ~= new FSMState!(T)();
			
			fsm[0].addNextState(opA.inclusivity, opA.inputs, fsm[1]);

			fsm[1].addNextState(RegexOpInclusivity.Include, opA.inputs, FSMState!(T).endState);
			fsm[1].addNextState(RegexOpInclusivity.Exclude, opA.inputs, FSMState!(T).endState);
			break;
			
		default:
			throw new Exception("FSM(T).this(RegexOp!(T), RegexOp!(T)): Unknown RegexOpMultiplicity");
		}
	}
*/	
	this(FSMState!(T) initState)
	{
		mixin(initMember!(initState));
		this.currState = initState;
	}
	
	private this(FSM!(T) copyOf)
	{
		mixin(initMemberFrom!(copyOf, initState, currState));
	}
	
	FSM!(T) dup()
	{
		return new FSM!(T)(this);
	}
	
	FSMState!(T) getState()
	{
		return currState;
	}
	
	void reset()
	{
		currState = initState;
	}
	
	void advance(T input)
	{
		currState = currState.getNextState(input);
	}
	
	bool failed()
	{
		return (currState is null);
	}
	
	bool succeeded()
	{
		return (currState is FSMState!(T).endState);
	}
	
	bool done()
	{
		return (failed() || succeeded());
	}
	
	static void output(FSMState!(T)[] states)
	{
		Stdout.formatln("Outputting FSM:");
		mixin(traceVal!("RegexOpInclusivity.Include"));
		mixin(traceVal!("states.length"));
		Stdout.formatln("");
		foreach(size_t i, FSMState!(T) state; states)
		{
			mixin(traceVal!("i"));
			state.output();
			Stdout.formatln("");
		}
	}
}

class FSMState(T)
{
	static const FSMState!(T) endState;
	static this()
	{
		endState = new FSMState!(T);
		endState.thisStatesIndex = size_t.max;
	}
	
	size_t thisStatesIndex;
	
	private FSMState!(T)[T[]][RegexOpInclusivity] nextStates;
//	public FSMAction defaultAction;
	
/*	public FSMAction getAction(T input)
	{
		if(input in actions)
			return actions[input];
		
		return defaultAction;
	}
*/

	void output()
	{
//		mixin(traceVal!("thisStatesIndex"));
//		mixin(traceVal!("nextStates.length"));
		foreach(RegexOpInclusivity incl, FSMState!(T)[T[]] nextStatesByInclusivity; nextStates)
		{
			mixin(traceVal!("incl"));
			foreach(T[] inputs, FSMState!(T) nextState; nextStatesByInclusivity)
			{
				mixin(traceVal!("inputs.length"));
				Stdout.formatln("inputs: '{}'", inputs);
				foreach(size_t i, T input; inputs)
					Stdout.formatln("inputs[{}]: 0x{:X2}", i, cast(byte)input);
				mixin(traceVal!("nextState.thisStatesIndex"));
			}
		}
	}
	
	FSMState!(T) getNextState(T input)
	{
		// Workaround for DMD Bugzilla #2192
		FSMState!(T) ret = null;
		
		foreach(RegexOpInclusivity lookupType, FSMState!(T)[T[]] nextStateList; nextStates)
		foreach(T[] inputs, FSMState!(T) state; nextStateList)
		{
			// Think of != as XOR
			if(tango.text.Util.contains(inputs, input) != (lookupType == RegexOpInclusivity.Exclude))
			{
				// Workaround for DMD Bugzilla #2192
				if(ret is null)
					ret = state;//return state;
				break;
			}
		}

		return ret;
	}
	
	bool pointsTo(FSMState!(T) state)
	{
		foreach(FSMState!(T)[T[]] nextStateList; nextStates)
		foreach(FSMState!(T) nextState; nextStateList)
		{
			if(nextState is state)
				return true;
		}

		return false;
	}
	
	void addNextState(RegexOp!(T) op, FSMState!(T) nextState)
	{
		mixin(traceVal!("op.inputs"));
//		mixin(traceVal!("op.inclusivity", "op.multiplicity", "nextState is endState", "nextState is null"));
		nextStates[op.inclusivity][op.inputs] = nextState;
	}
	
	void addNextState(RegexOpInclusivity lookupType, T[] inputs, FSMState!(T) nextState)
	{
		mixin(traceVal!("inputs"));
		nextStates[lookupType][inputs] = nextState;
	}
	
	void addNextStateIf(T[] inputs, FSMState!(T) nextState)
	{
		addNextState(RegexOpInclusivity.Include, inputs, nextState);
	}
	
	void addNextStateIfNot(T[] inputs, FSMState!(T) nextState)
	{
		addNextState(RegexOpInclusivity.Exclude, inputs, nextState);
	}
}

//TODO: Create validateFSM(initialFSMState)

/**
Generates an FSM from a psuedo-regex string.

Psuedo-Regex String:
	Psuedo-regex strings are case-sensitive.
	
	Using non-escaped Special Characters where they aren't
	expected is invalid and will trigger an Exception.
	
	When creating a psuedo-regex string literal, using D's
	WYSIWYG	strings (ie, r"" or ``) is highly recommended.
	
	Special Characters:
		\ [ ] + * . ^ newline(any OS format) tab verticaltab formfeed
	
	Features:
	(Assume "String" and "Character" don't contain any Special Characters.)
	String     : Match sequence of characters exactly.
	[String]   : Match any single character in String.
	[^String]  : Match any single character NOT in String.
	.          : Matches any single character.
	Character+ : Matches one or more instance of Character.
	Character* : Matches zero or more instances of Character.
	\Character : Escape sequence:
				 \\ : Match backslash
				 \[ : Match opening square bracket
				 \] : Match closing square bracket
				 \+ : Match plus sign
				 \* : Match asterisk
				 \. : Match period
				 \^ : Match caret
				 \d : Match any digit 0 through 9
				 \a : Match any lowercase english letter
				 \A : Match any uppercase english letter
				 \w : Word Character, same as r"[\d\a\A_]"
				 \t : Match tab
				 \v : Match vertical tab (rarely used)
				 \f : Match form feed (rarely used)
				 \n : Match newline in any OS format: Win, Mac, or Unix
				 \s : Whitespace, same as r"[ \t\v\f\n]" 
				 Anything else (including end of string) after the '\' is
				 invalid and triggers an Exception.
	All combinations of the above features are allowed.
	No features besides those above are supported.
	
	Examples:
	r"ABC"    : Matches r"ABC" (and nothing else).
	r"[ABC]"  : Matches r"A", r"B", or r"C" (and nothing else).
	r"A[BC]D" : Matches r"ABC", or r"ACD" (and nothing else).
	r"[A]+"   : Matches one or more consecutive r"A".
	r"[^A]"   : Matches any single-character string except r"A".
	r"[^AB]"  : Matches any single-character string except r"A" and r"B".
	
	r"[\d]+"        : Matches positive integers.
	r"[\a\A_][\w]*" : Matches C-style variable names.
	r"//[^\n]*"     : Matches C-style EOL comments.

Returns:
	Initial state of FSM
*/
/*LexFSMState genFSM(schar[] str)
{
	LexFSMState initialState = new LexFSMState();
	
	
	return initialState;
}
*/

/// Unused doc comment snippet:
	/**
	Paramater 'inputs':
		Case-sensitive.
		Use of D's WYSIWYG strings is recommended for literals.
		
		Examples:
		"+XYZ" : Go to nextState if input is 'X', 'Y', or 'Z'
		"-XYZ" : Go to nextState if input is NOT 'X', 'Y', or 'Z'
		"+XY\d": Go to nextState if input is 'X', 'Y', or a digit '0' through '9'
		"+7\l" : Go to nextState if input is '7' or a lowercase english letter
		"+7\u" : Go to nextState if input is '7' or an uppercase english letter
		"+\n"  : Go to nextState if input is a newline (any OS format)
		"++\s" : Go to nextState if input is '+' or whitespace (including newline)
		"+ -"  : Go to nextState if input is '-' or a space
		"-.\\" : Go to nextState if input is neither '.' nor a backslash
		"+"    : There's nothing that causes FSM to go to nextState
				 (This is disallowed and triggers an exception.)
		"-"    : There's nothing that prevents the FSM from going to nextState.
				 Ie, go to nextState regardless of input.
				 (disallowed just for now)
	*/
