// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.lang.gold;

import tango.io.Stdout;
import tango.text.Regex;
import tango.text.Util;

import semitwist.util;
import semitwist.treeout;

import stc.util;
import stc.fsm;
import stc.language;

import stc.goldbase;
import stc.goldfile;

public class TokGold : Token
{
	private GoldSymbolType _type;
	private GoldSymbol _symbol;
	char[] _file;
	int _line;
	int _srcIndexStart;
	int _srcIndexEnd;
	
	//TODO: Expose as opIndex
	private TokGold[] subTokens;
	
	public this(GoldSymbol symbol, schar[] content, char[] file="{unknown}", int line=1, int srcIndexStart=0, int srcIndexEnd=0)
	{
		super(content);
		this._symbol = symbol;
		this._type = cast(GoldSymbolType)symbol.type;
		this._file = file;
		this._line = line;
		this._srcIndexStart = srcIndexStart;
		this._srcIndexEnd   = srcIndexEnd;
		
		if(_type == GoldSymbolType.NonTerminal)
			throw new Exception("Called wrong constructor for NonTerminal");
	}
	
	public this(GoldSymbol symbol, TokGold[] subTokens)
	{
		super("");
		this._symbol = symbol;
		this._type = cast(GoldSymbolType)symbol.type;
		mixin(initMember!(subTokens));
		
		if(_type != GoldSymbolType.NonTerminal)
			throw new Exception("Called wrong constructor for token that isn't a NonTerminal");
	}
	
	public TokGold getFirstLeaf()
	{
		if(_type == GoldSymbolType.NonTerminal && subTokens.length > 0)
		{
			foreach(TokGold subToken; subTokens)
			{
				if(subToken.symbolType() != GoldSymbolType.NonTerminal || subToken.subTokens.length > 0)
					return subToken.getFirstLeaf();
			}
		}

		return this;
	}
	
	public TokGold getLastLeaf()
	{
		if(_type == GoldSymbolType.NonTerminal && subTokens.length > 0)
		{
			foreach_reverse(TokGold subToken; subTokens)
			{
				if(subToken.symbolType() != GoldSymbolType.NonTerminal || subToken.subTokens.length > 0)
					return subToken.getLastLeaf();
			}
		}

		return this;
	}
	
	public int srcIndexStart()
	{
		if(_type == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getFirstLeaf().srcIndexStart();
		else
			return _srcIndexStart;
	}
	
	public int srcIndexEnd()
	{
		if(_type == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getLastLeaf().srcIndexEnd();
		else
			return _srcIndexEnd;
	}
	
	public int line()
	{
		if(_type == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getFirstLeaf().line();
		else
			return _line;
	}
	
	public char[] file()
	{
		if(_type == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getFirstLeaf().file();
		else
			return _file;
	}
	
	public GoldSymbolType symbolType()
	{
		return _type;
	}
	
	public char[] symbolTypeName()
	{
		return goldSymbolTypeToString(_type);
	}

	public wchar[] symbolName()
	{
		return _symbol.name.dup;
	}

	public wchar[] symbolFullName()
	{
		return stformatw("{}.{}", symbolTypeName(), symbolName());
	}
	
	public int symbolId()
	{
		return _symbol.id;
	}

	//TODO: Can I do toString(bool raw=false)? (Ie, will Stdout.formatln("{}", this) still work?)
	public char[] toString()
	{
		char[] ret;
		
		switch(_type)
		{
		case GoldSymbolType.NonTerminal:
			foreach(TokGold tok; subTokens)
				ret ~= tok.toString();
			break;
			
		case GoldSymbolType.Terminal:
			ret = content.dup;
			break;
			
		case GoldSymbolType.Whitespace:
		case GoldSymbolType.EOF:
		case GoldSymbolType.CommentStart:
		case GoldSymbolType.CommentEnd:
		case GoldSymbolType.CommentLine:
		case GoldSymbolType.Error:
			ret = "";
			break;
			
		default:
			throw new Exception(stformat("Unknown GoldSymbolType: #{}", _type));
		}
		
		return ret;
	}
	
	public char[] toStringRaw()
	{
		char[] ret;
		
		if(_type == GoldSymbolType.NonTerminal)
		{
			foreach(TokGold tok; subTokens)
				ret ~= tok.toString();
		}
		else
			ret = content.dup;
			
		return ret;
	}
	
	public TreeNode toTreeNode()
	{
		auto node = new TreeNode( "{}".stformat(symbolName()) );
		node.addAttribute("srcIndexStart", srcIndexStart());
		node.addAttribute("srcIndexEnd", srcIndexEnd());
		node.addAttribute("line", line());
		node.addAttribute("type", symbolTypeName());
		node.addAttribute("symbol", symbolName());
		node.addAttribute("content", toStringRaw());
		
		//node.addContent(new XmlNode("content", new XmlNodeCData(toStringRaw())));
		
		if(_type == GoldSymbolType.NonTerminal)
		foreach(TokGold tok; subTokens)
			node.addContent(tok.toTreeNode());
		
		return node;
	}
	
	public TreeNode toPsuedoTreeNode()
	{
		auto node = new TreeNode(stformat("{}", toStringRaw()));
		node.addAttribute("symbol", symbolFullName());

		if(_type == GoldSymbolType.NonTerminal)
		foreach(TokGold tok; subTokens)
			node.addContent(tok.toPsuedoTreeNode());
		
		return node;
	}
}

public class GoldLanguage : Language
{
	// Think about turning these into accessors,
	// or just move the lexing/parsing into this class
	public wchar[] grammarName;
	public wchar[] grammarVersion;
	public wchar[] grammarAuthor;
	public wchar[] about;
	public bool    caseSensitive;
	
	public GoldSymbol[]     symbolTable;
	public GoldCharSet[]    characterSetTable;
	public GoldRule[]       ruleTable;
	public GoldDFAState[]   dfaTable;
	public GoldLALRState[]  lalrTable;
	
	public int startSymbolIndex;
	public int initialDFAState;
	public int initialLALRState;
	
	public int eofSymbolIndex;
	public int errorSymbolIndex;

	public this(char[] cgtFilename)
	{
		auto goldFile = new GoldFile(cgtFilename);
		
		mixin(initMemberFrom!(goldFile,
		                      grammarName, grammarVersion, grammarAuthor,
							  about, caseSensitive,
		                      startSymbolIndex, initialDFAState, initialLALRState,
	                          symbolTable, characterSetTable, ruleTable,
						      dfaTable, lalrTable));

		eofSymbolIndex   = getSymbolIndexByType(GoldSymbolType.EOF);
		errorSymbolIndex = getSymbolIndexByType(GoldSymbolType.Error);

		mixin(traceVal!("grammarName", "grammarVersion", "grammarAuthor",
		                "about", "caseSensitive",
						"startSymbolIndex", "initialDFAState", "initialLALRState", 
						"symbolTable.length", "characterSetTable.length",
						"ruleTable.length", "dfaTable.length", "lalrTable.length"
						));
		mixin(traceVal!("eofSymbolIndex", "errorSymbolIndex"));
		mixin(traceVal!("symbolTable[eofSymbolIndex]", "symbolTable[errorSymbolIndex]"));
		mixin(traceVal!("symbolTable"));
	/*	Stdout.formatln("");
		mixin(traceVal!("symbolTable", "characterSetTable",
						"ruleTable", "dfaTable", "lalrTable"
						));
	*/					
	}

	private int getSymbolIndexByType(GoldSymbolType type)
	{
		foreach(int i, GoldSymbol symbol; symbolTable)
			if(symbol.type == type)
				return i;
		throw new Exception("No symbol found by getSymbolIndexByType for type {}".stformat(type));
	}
	
	public GoldSymbol eofSymbol()
	{
		return symbolTable[eofSymbolIndex];
	}
	
	public GoldSymbol errorSymbol()
	{
		return symbolTable[errorSymbolIndex];
	}
	
	public TokenType[] getTokenTypes()
	{
		TokenType[] tokTypes;
		
		tokTypes ~= new TokenType(r":==",
						delegate Token(schar[] data)
						{
							throw new Exception(stformat("(!!!{}!!!USED TOKENTYPE!!!)", data));
						});

		return tokTypes;
	}
}
