// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.lexer;

import tango.core.Array;
import tango.io.Stdout;
import tango.text.Regex;
import util = tango.text.Util;

import semitwist.util;

import stc.util;
import stc.fsm;
import stc.parser;
import stc.language;
import stc.goldbase;
import stc.goldfile;

import stc.lang.gold;
import stc.lang.stca;

alias FSM!(schar) LexFSM;
alias FSMState!(schar) LexFSMState;

class Lexer
{
	// For debugging
	schar[] remainingData;
	schar[] processedData;
	schar[] mixedData;
	
	private LexFSM[] fsms;
	private RegExpT!(schar)[] regexes;
	private TokenType[] tokTypes;
	private uint numErrors;

	private void AppendRemainingData(T)(T data)
	{
		remainingData ~= data;
		mixedData ~= data;
	}

	private void AppendProcessedData(T)(T data)
	{
		processedData ~= data;
		mixedData ~= data;
	}
	
	uint getNumErrors()
	{
		return numErrors;
	}
	
	private Token[] process(schar[] source, char[] filename, GoldLanguage lang)
	{
		TokGold[] tokens;
		schar[] tok;
		schar[] errTok;
		
		int srcIndex=0;
		int line=1;
		int lineOfErr;
		int state = lang.initialDFAState;
		bool eof = false;
		
		//TODO: Expose this
		int[] lineIndicies;
	
		bool advanceState(schar c)
		{
			bool charAccepted = false;
			foreach(GoldDFAStateEdge edge; lang.dfaTable[state].edges)
			{
				if(lang.characterSetTable[edge.charSetIndex].chars.contains(c))
				{
					state = edge.targetDFAStateIndex;
					charAccepted = true;
					tok ~= c;
					break;
				}
			}
			return charAccepted;
		}
		
		void addToken(GoldSymbol symbol, schar[] content)
		{
			if(errTok.length > 0)
			{
				Stdout.formatln("{}(index {}): Syntax Error: '{}'",
								filename, srcIndex, errTok);
				numErrors++;
				tokens ~= new TokGold(lang.errorSymbol(), errTok, filename, lineOfErr, srcIndex-errTok.length, srcIndex);
				AppendRemainingData(errTok);
				errTok = "";
			}

			tokens ~= new TokGold(symbol, content, filename, line, srcIndex-content.length, srcIndex);
			if(symbol.type == GoldSymbolType.Whitespace)
				AppendProcessedData(stformat("{}", tok));
			else
				AppendProcessedData(stformat("[:{}:]", tok));
		}
		
		bool isNewline()
		{
			bool ret=false;
			if(srcIndex < source.length)
			{
				if(source[srcIndex] == '\n')  // Win and Unix style
					ret = true;
				else
				{
					if(srcIndex+1 < source.length)
					{
						// Mac style
						if(source[srcIndex] == '\r' && source[srcIndex+1] != '\n')
							ret = true;
					}
				}
			}
			return ret;
		}
		
		lineIndicies ~= 0;
		if(source.length != 0)
		while(!eof)
		{
			schar c;
			bool charAccepted = false;

			if(srcIndex == source.length)
				eof = true;
				
			if(!eof)
			{
				c = source[srcIndex];
				charAccepted = advanceState(c);
			}
			
			bool shouldAdvance = true;
			if(!charAccepted)
			{
				if(lang.dfaTable[state].accept)
				{
					auto symbolIndex = lang.dfaTable[state].acceptSymbolIndex;
					addToken(lang.symbolTable[symbolIndex], tok);
					shouldAdvance = false;
				}
				else
				{
					if(errTok == "")
						lineOfErr = line;
						
					errTok ~= c;
				}

				state = lang.initialDFAState;
				tok = "";
			}
			
			if(shouldAdvance)
			{
				srcIndex++;
					
				if(isNewline())
				{
					line++;
					
					if(srcIndex+1 < source.length)
						lineIndicies ~= srcIndex+1;
				}
			}
		}
		
		addToken(lang.eofSymbol(), "{EOF}");
		return tokens;
	}
	
	//TODO: Handle EOF
	Token[] process(schar[] source, char[] filename, Language lang)
	{
		remainingData = processedData = mixedData = "";

		if(cast(GoldLanguage)lang)
		{
			return process(source, filename, cast(GoldLanguage)lang);
		}
		else // Only GoldLanguage supported at the moment
		{
			Token[] dummy;
			return dummy;
		}
/+
		remainingData = emptyString!(schar)();
		processedData = emptyString!(schar)();
		mixedData     = emptyString!(schar)();
		
		Token[] tokens;

		size_t lineNumber=1;
		size_t startOfToken=0;  // Probably more like "Place to start next token if this one fails"
		schar[] currLine;
		schar[] currToken;
		
		// Standardize line endings to "\n"
		toUnixEOL(source);

		tokTypes = lang.getTokenTypes();
		foreach(TokenType tt; tokTypes)
		{
			fsms ~= tt.getFSM();
			regexes ~= tt.getRegex();
		}
		
		if(fsms[0] is null)
		{
			foreach(RegExpT!(schar) r; regexes)
				r.test(source);
			
			size_t[] indicies;
			bool[] matchFound;
			indicies.length = regexes.length;
			matchFound.length = regexes.length;

			size_t nearestMatch;

			foreach(size_t indexR, RegExpT!(schar) r; regexes)
			{
				matchFound[indexR] = regexes[indexR].test(source);
				if(matchFound[indexR])
					indicies[indexR] = regexes[indexR].pre().length;
				else
					indicies[indexR] = size_t.max;
			}

			size_t startingPoint=0;
			bool done=false;
			while(!done)
			{
				nearestMatch = indexOfMin(indicies);
				if(matchFound[nearestMatch])
				{
					// Lex Error
					if(indicies[nearestMatch] > startingPoint)
					{
						numErrors++;
						
						schar[] invalidData = source[startingPoint..indicies[nearestMatch]];
						AppendRemainingData(invalidData);

/*						Stdout.formatln("{0}({1}): Syntax Error: char(0x{3:X2}) '{4}' \"{}\"",
										filename, lineNumber, currLine.length, cast(uint)c, c, currToken);
						Stdout.formatln("{}", currLine);
						Stdout.formatln("{}^", repeat("-", currLine == "" ? 0 : currLine.length-1));
*/
						Stdout.formatln(`{0}({1}): Syntax Error: "{}"`,
										filename, lineNumber, invalidData);
					}
				
					// Emit
					AppendProcessedData("[:"~regexes[nearestMatch][0]~":]");
					tokens ~= tokTypes[nearestMatch].newToken(regexes[nearestMatch][0]);
					
					// Find Next Token
					startingPoint = indicies[nearestMatch] + regexes[nearestMatch][0].length;
					foreach(size_t indexR, RegExpT!(schar) r; regexes)
					{
						while(indicies[indexR] < startingPoint && matchFound[indexR])
						{
							matchFound[indexR] = regexes[indexR].test();
							if(matchFound[indexR])
								indicies[indexR] = regexes[indexR].pre().length;
							else
								indicies[indexR] = size_t.max;
						}
					}
				}
				else
					done = true;
			}
		}
		else
		{

LexFSMState junk = null;
mixin(traceVal!("fsms[0].getState()", "junk", "fsms[0].getState() is null",
"fsms[0].getState() is LexFSMState.endState", "LexFSMState.endState is null",
"fsms[0].getState() is fsms[1].getState()",
"fsms[0].getState() is fsms[2].getState()",
"fsms[0].getState() is fsms[3].getState()",
"fsms[0].getState() is fsms[4].getState()"//,
//"fsms[0].getState() is fsms[5].getState()"
));
		

		for(size_t i=0; i < source.length; i++)
		{
			schar c = source[i];
//			if(i == 3)
//				break;
			
			if(c == '\n')
			{
				lineNumber++;
				currLine = "";
			}
			else
				currLine ~= c;
				
			currToken ~= c;
//			mixin(traceVal!("c"));
//			mixin(traceVal!("i"));
			
			size_t numFailed=0;
			size_t numEnded=0;
			size_t newlyEndedFSMIndex;
			LexFSM newlyEndedFSM;
			foreach(size_t fsmIndex, LexFSM fsm; fsms)
			{
//				mixin(traceVal!("        fsmIndex"));
				if(fsm.failed())
				{
					numFailed++;
					numEnded++;
				}
				else if(fsm.succeeded())
				{
					numEnded++;
				}
				else
				{
					fsms[fsmIndex].advance(c);
					if(fsms[fsmIndex].failed())
					{
						numFailed++;
						numEnded++;
					}
					else if(fsms[fsmIndex].succeeded())
					{
						numEnded++;
						newlyEndedFSMIndex = fsmIndex;
						newlyEndedFSM = fsm;
					}
				}
//				mixin(traceVal!("numFailed", "numEnded"));
			}
			
			if(numEnded == fsms.length)
			{
				if(numFailed == fsms.length)
				{
					// Lex error

					// Why the hell does this seem to cause lexing work?
					if(i != startOfToken)
					{
						currToken = currToken[0..$-1];
						if(currLine.length > 0)
							currLine = currLine[0..$-1];
					}
					
					Stdout.formatln("{0}({1}): Syntax Error: char(0x{3:X2}) '{4}' \"{}\"",
									filename, lineNumber, currLine.length, cast(uint)c, c, currToken);
					Stdout.formatln("{}", currLine);
					Stdout.formatln("{}^", util.repeat("-", currLine == "" ? 0 : currLine.length-1));
					
					AppendRemainingData(currToken);
					
					i = startOfToken;
					startOfToken++;
				}
				else
				{
					currToken = currToken[0..$-1];
					if(currLine.length > 0)
						currLine = currLine[0..$-1];

					// Emit
					AppendProcessedData("[:"~currToken~":]");
					tokens ~= tokTypes[newlyEndedFSMIndex].newToken(currToken);
						
					startOfToken = i;
					i--;
				}

				// Reset
				foreach(LexFSM fsm; fsms)
					fsm.reset();
				currToken = "";
			}
		}
		}
		
		return tokens;
+/
	}
}
