// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.main;

import tango.core.Exception;
import tango.io.Stdout;
//import tango.io.Fileconduit;
import tango.io.device.Fileconduit;
import tango.text.Util;
import tango.text.convert.Integer;
import layout = tango.text.convert.Layout;

//import mtext;
import semitwist.util;
import semitwist.cmdlineparser;
//import semitwist.xmlout;
import semitwist.treeout;

import stc.util;
import stc.lexer;
import stc.parser;
import stc.language;
//import stc.grammar;
import stc.lang.gold;
import stc.lang.stca;

import stc.goldbase;
import stc.goldfile;

bool htmlOutput=false;
TreeNodeBase[] htmlOutBuffer;

//TODO: Make commandline args, including "-uv": Ultra-Verbose
void main(char[][] args)
{
	scope(exit) if(htmlOutput) displayHtmlOutput(false);
	
	if(getNonFatalAssertCount() > 0)
	{
		Stdout.flush();
		assert(false, stformat("{} Assert Failure{}",
		                       getNonFatalAssertCount(),
					           getNonFatalAssertCount() == 1 ? "" : "s")
		      );
	}
	outputHeader();
   
	bool help;
	bool detailhelp;
	char[] lang = "GOLD";
	char[] gold;// = "webscript.cgt";
	char[] src;
	char[] parseTreeOut;
	char[] parseTreeType="JSON";

	bool saveParseTree=false;
	bool parseTreePretty=false;
	
	auto cmd = new CmdLineParser();
	mixin(defineArg!(cmd, "help",   help,            ArgFlag.Optional,   "Displays a help summary and exits" ));
	mixin(defineArg!(cmd, "detail", detailhelp,      ArgFlag.Optional,   "Displays a detailed help message and exits" ));
	mixin(defineArg!(cmd, "html",   htmlOutput,      ArgFlag.Optional,   "Make compiler's stdout HTML instead of text (experimental)" ));
	mixin(defineArg!(cmd, "lang",   lang,            ArgFlag.Optional,   `Language: "GOLD" or "STCA"` ));
	mixin(defineArg!(cmd, "gold",   gold,            ArgFlag.Optional,   "GOLD Parser CGT file of the input language" ));
	mixin(defineArg!(cmd, "src",    src,             ArgFlag.Switchless, "Input filename" ));
	mixin(defineArg!(cmd, "pt",     saveParseTree,   ArgFlag.Optional,   "Save Parse Tree" ));
	mixin(defineArg!(cmd, "ptp",    parseTreePretty, ArgFlag.Optional,   `Pretty Format Parse Tree (Implies /pt)` ));
	mixin(defineArg!(cmd, "ptt",    parseTreeType,   ArgFlag.Optional,   `Parse Tree Type ("JSON" or "XML")` ));
	mixin(defineArg!(cmd, "pto",    parseTreeOut,    ArgFlag.Optional,   "Parse Tree Filename (Only outputs on successful parse)" ));
	
	if(!cmd.parse(args)
		|| help
		|| (!detailhelp && (src == "" || (parseTreeType != "JSON" && parseTreeType != "XML")))
	  )
	{
		displayln("");
		display(cmd.getUsage());
		return;
	}
	if(detailhelp)
	{
		displayln("");
		display(cmd.getDetailedUsage());
		return;
	}
	displayln("");
	
	TreeFormatter formatter = 
		(parseTreeType=="XML"  &&  parseTreePretty)? formatterPrettyXML :
		(parseTreeType=="XML"  && !parseTreePretty)? formatterTrimmedXML :
		(parseTreeType=="JSON" &&  parseTreePretty)? formatterPrettyJSON :
		formatterTrimmedJSON;
		
	char[] inputFilename = src;

	if(parseTreePretty)
		saveParseTree = true;
		
	if(parseTreeOut == "")
		parseTreeOut = inputFilename~".pt."~((parseTreeType=="XML")?"xml":"json");
		
	if(gold == "")
	{
		char[] infileExt = inputFilename[inputFilename.locatePrior('.')+1..$];
		gold = infileExt~".cgt";
	}

	//	testGrammar();

	displayln("Input File: "~inputFilename);
	flushDisplay();

	auto source = loadFile(inputFilename);
	
	if(lang == "GOLD")
	{
		auto goldFileName = gold;
		displayln("Gold Language: "~goldFileName);
		displayln("");
		flushDisplay();

		auto goldLang = new GoldLanguage(goldFileName);

		outputSection("Input", source);
		process(source, inputFilename, goldLang, saveParseTree, parseTreeOut, parseTreeType, formatter);

	}
	else if(lang == "STCA")
	{
		displayln("Language: STCA");
		displayln("");
		outputSection("Input", source);

		//process(source, inputFilename, new STCA());
	}
}

void process(schar[] source, char[] inputFilename, Language lang, bool saveParseTree, char[] parseTreeOut, char[] parseTreeType, TreeFormatter formatter)
{
	outputSection("Lexing");
	auto lexer = new Lexer();
	Token[] tokens;
	try
		tokens = lexer.process(source, inputFilename, lang);
	catch(Exception e)
		displayln(stformat("Lexer: {}", e));
	
	displayResults(lexer, tokens);

	if(cast(GoldLanguage)lang)
	{
		outputSection("Parsing");
		auto parser = new Parser();
		try
			tokens = parser.process(tokens, inputFilename, cast(GoldLanguage)lang);
		catch(Exception e)
		{
			displayln(stformat("Parser: {}", e));

			outputSection("Parser Info");
			displayResults(parser.tokenStack);
			mixin(traceVal!("parser.stateStack", "parser.rulesUsed"));
		}
		
		outputSection("Tokens");
		displayResults(tokens);
		
		if(tokens.length == 1)
		{
			displayln("Source Parsed Successfully!");

			if(saveParseTree)
			{
				
				auto parseTreeRoot = (cast(TokGold)tokens[0]).toTreeNode();
				parseTreeRoot.addAttribute("file", inputFilename);
				char[] parseTreeStr =
					parseTreeRoot
					.format(formatter)
					.toNativeEOLFromUnix();
					
				//outputSection("Parse Tree");
				//displayln( parseTreeStr );
				display("Saving {} Parse Tree ({})...".stformat(parseTreeType, parseTreeOut));
				saveFile(parseTreeOut, cast(ubyte[])parseTreeStr);
				displayln("done");
			}
		}
		else
			displayln("Source Failed to Parse!");
	}
}

void displayResults(Lexer lexer, Token[] tokens)
{
	displayln(stformat("{} error{}",
					lexer.getNumErrors(),
					lexer.getNumErrors() == 1 ? "" : "s"));
	
	outputSection("Remaining Data", lexer.remainingData.toNativeEOLFromUnix());
	outputSection("Processed Data", lexer.processedData.toNativeEOLFromUnix());
	outputSection("Mixed Data",     lexer.mixedData.toNativeEOLFromUnix());
	outputSection("Tokens");

	displayResults(tokens);

	if(lexer.remainingData == "")
		displayln(stformat("File lexed completely!"));
}

void displayResults(Token[] tokens)
{
	bool gold=false;
	foreach(int i, Token tok; tokens)
	{
		TokGold tokGold = cast(TokGold)tok;
		if(tokGold)
		{
			gold = true;
			if(tokGold.symbolType() != GoldSymbolType.Whitespace)
				displayln(stformat("{}({}): [{}]{,-25} '{}'",
				                tokGold.file(), tokGold.line(), i,
				                tokGold.symbolFullName() ~ ":",
								tokGold.toStringRaw()));

		}
		else
		{
//			if(!cast(STCA.TokWhitespace)tok)
//				displayln(stformat("{,-35} '{}'", tok.toString ~ ":", tok.getContent()));
		}
	}
	
	// ------------
	
/*	if(gold)
	{
		XmlNode[] xmlTokens;
		foreach(Token tok; tokens)
		{
			auto tokGold = cast(TokGold)tok;
			
			auto xmlTok = new XmlNode("token", new XmlNodeData(tokGold.toStringRaw()));
			xmlTok.addAttribute("line", tokGold.line());
			xmlTok.addAttribute("symbolName", tokGold.symbolFullName());

			xmlTokens ~= xmlTok;
		}
		outputSection( "XML Token Data",
					   (new XmlNode("tokens", xmlTokens)).toString(false).toNativeEOLFromUnix() );
	}
*/
}

void outputHeader()
{
	displayln("");
	displayln(stformat("SemiTwist Compiler"));
}

void outputSection(T,U=void)(T[] heading, U[] content=null)
{
	displayln(stformat("------- {}:", heading));
	if(content != null)
		displayln(stformat("{}", content));
}

/*void testGrammar()
{
	auto test = new GrammarTerminal("My Token Rule Name", "Some Pattern");
	mixin(traceVal!("test.getName()", "test.getPattern()"));
	mixin(traceVal!("test"));
	Stdout.formatln("");
	
	auto testTerminal = new GrammarTerminal("TestTerminal", "TestTerminal Pattern");
	
	auto testClauseNull = new GrammarClause(null);
	auto testNonTerminalNull = new GrammarNonTerminal("TestNonTerminalNull", null);
	
	auto testClause = new GrammarClause([cast(GrammarNode)testTerminal, testNonTerminalNull, testTerminal]);
	auto testNonTerminal = new GrammarNonTerminal("TestNonTerminal", [testClauseNull, testClause]);
												
	mixin(traceVal!("testTerminal",
					"testClauseNull", "testNonTerminalNull",
					"testClause",     "testNonTerminal"));
}
*/
schar[] loadFile(char[] filename)
{
	auto inFile = new FileConduit(filename);
	auto inFileLength = inFile.length;
	//WARNING: file can't be larger than uint.max
	//TODO: Convert lex/parse to work as producers/consumers, like in GOLD docs,
	//      instead of all fully completing each step before next.
	schar[] inFileContent = cast(schar[])(new byte[cast(uint)inFileLength]);
	auto inFileLengthRead = inFile.input.read(inFileContent);
	inFile.close();
	
	if(inFileLengthRead != inFileLength)
		throw new Exception(stformat("Failed to read entire file: {}: Read {} out of {} bytes.",
									  filename, inFileLengthRead, inFileLength));

	return inFileContent;
}

void saveFile(char[] filename, ubyte[] data)
{
	auto outFile = new FileConduit(filename, FileConduit.WriteCreate);
	auto bytesWritten = outFile.output.write(data);
	outFile.close();
	
	if(bytesWritten != data.length)
		throw new Exception(stformat("Failed to write entire file: {}: Wrote {} out of {} bytes.",
									  filename, bytesWritten, data.length));
}

void display(char[] str)
{
	if(htmlOutput)
		htmlOutBuffer ~= new TreeNodeData(str);
	else
		Stdout.format("{}", str);
}

void displayln(char[] str)
{
	if(htmlOutput)
		display(str~r"<br \>");
	else
		display(str~nativeEOL);
}

void flushDisplay()
{
	if(htmlOutput)
	{
		foreach(TreeNodeBase node; htmlOutBuffer)
			Stdout.formatln("{}", node.toString(formatterTrimmedXML, "", 0).toNativeEOLFromUnix());
	}
	Stdout.flush();
}

void displayHtmlOutput(bool strip=true)
{
	auto htmlBody  = new TreeNode("body", htmlOutBuffer);
	auto htmlTitle = new TreeNode("title", new TreeNodeData("SemiTwist Compiler Output"));
	auto htmlHead  = new TreeNode("head", htmlTitle);
	auto htmlRoot  = new TreeNode("html");
	htmlRoot.subNodes ~= htmlHead;
	htmlRoot.subNodes ~= htmlBody;
	Stdout.formatln("{}", htmlRoot.format(formatterTrimmedXML).toNativeEOLFromUnix());
}
