// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.parser;

import tango.io.Stdout;
//import tango.text.Regex;
import tango.text.Util;

import semitwist.util;

import stc.util;
import stc.fsm;
import stc.language;

import stc.goldbase;
import stc.lang.gold;


class Parser
{
	int[] rulesUsed;
	TokGold[] tokenStack;
	int[] stateStack;
	
	Token[] process(Token[] tokens, char[] filename, GoldLanguage lang)
	{
		rulesUsed.length = 0;

		int tokIndex=0;
		int state=lang.initialLALRState;
		bool done=false;
		
		TokGold getPendingToken()
		{
			//TODO: Alter this to skip past comments (not just whitespace)
			TokGold tok;
			bool inComment=false;
			bool lineComment;
			do
			{
				tok = cast(TokGold)tokens[tokIndex];
				if(!tok)
					throw new Exception(stformat("Internal Error: Non-TokGold found in token list: #{}", tokIndex));
				tokIndex++;
				
				// Handle comments
				//TODO: Move this into the lexer, because of awkwardness of handling newlines here
				if(!inComment)
				{
					switch(tok.symbolType())
					{
					case GoldSymbolType.CommentLine:
						inComment = true;
						lineComment = true;
						break;
					case GoldSymbolType.CommentStart:
						inComment = true;
						lineComment = false;
						break;
					default:
						break;
					}
				}
				else
				{
					if(lineComment)
					{
						char[] tokContent = tok.toStringRaw();
						if(tokContent.contains('\n') || tokContent.contains('\r'))
							inComment = false;
					}
					else
					{
						if(tok.symbolType() == GoldSymbolType.CommentEnd)
							inComment = false;
					}
				}
			}
			while(inComment ||
			         (
				         tok.symbolType() != GoldSymbolType.Terminal &&
			             tok.symbolType() != GoldSymbolType.EOF
			         )
			     );
			tokIndex--;
			return tok;
		}

		void shift()
		{
			if(tokIndex >= tokens.length)
				throw new Exception("Unexpected EOF");
				
			tokenStack ~= getPendingToken();
			stateStack ~= state;
			tokIndex++;
		}
		
		GoldLALRAction chosenAction;
		int chosenActionId;
		bool actionFound;

		void lookupAction(TokGold tok)
		{
			actionFound=false;
			foreach(int id, GoldLALRAction action; lang.lalrTable[state].actions)
			{
				if(tok.symbolId() == action.symbolId)
				{
					chosenAction = action;
					chosenActionId = id;
					actionFound = true;
					break;
				}
			}
			//mixin(traceVal!("tok", "tok.symbolId()", "actionFound", "chosenActionId", "chosenAction"));
			//mixin(traceVal!("actionFound", "chosenAction.type"));
		}
		
		void displayStack()
		{
/*			foreach(TokGold tok; tokenStack)
			{
				Stdout.formatln("    {}({}): {,-25} '{}'",
								tok.file(), tok.line(),
								tok.symbolFullName() ~ ":",
								tok.toStringRaw());
			}
*/		}
		
		while(!done)
		{
			auto tok = getPendingToken();
			lookupAction(tok);
			
			if(actionFound)
			{
				switch(chosenAction.type)
				{
				case GoldLALRAction.Type.Shift:
					shift();
					state = chosenAction.target;
					//mixin(traceVal!("chosenActionId"));
					//Stdout.formatln(`Shift:  {,-10} new state: {}`,
					//                `"`~tok.toStringRaw()~`"`, state);
					displayStack();
					break;

				case GoldLALRAction.Type.Reduce:
					auto rule = lang.ruleTable[chosenAction.target];
					auto reductionSize = rule.terminalIndicies.length;
					if(tokenStack.length < reductionSize)
						throw new Exception(
							stformat("Internal Error: Tried to reduce more tokens than exist on the stack. Have {}, need {}, rule #{}",
									 tokenStack.length, reductionSize, chosenAction.target));
					rulesUsed ~= chosenAction.target;
					
					// Perform Reduction
					TokGold[] reducedTokens;
					if(reductionSize > 0)
					{
						auto reductionStart = tokenStack.length - reductionSize;
						reducedTokens = tokenStack[reductionStart..$].dup;
						tokenStack = tokenStack[0..reductionStart];
						state = stateStack[reductionStart];
						stateStack = stateStack[0..reductionStart];
					}

					auto reducedToken = new TokGold(lang.symbolTable[rule.nonTerminalIndex], reducedTokens);
					tokenStack ~= reducedToken;
					stateStack ~= state;
					
					// Ensure reduced tokens symbols match the rule's right-side
					if(reducedTokens.length != reductionSize)
						throw new Exception(
							stformat("Internal Error: Reduced an incorrent number of tokens. Expected {}, Actual {}",
							         reductionSize, reducedTokens.length));
						
					foreach(int i, TokGold t; reducedTokens)
					{
						if(t.symbolId() != rule.terminalIndicies[i])
							throw new Exception(
								stformat(`Internal Error: Reduced token #{} does not match rule. Expected "{}", Actual "{}"`,
								         i, lang.symbolTable[rule.terminalIndicies[i]].name, t.symbolName()));
					}
					//mixin(traceVal!("chosenActionId"));
					//Stdout.format(`Reduce: {,-15} (next: "{}") new state: {}`,
					//              `"`~reducedToken.symbolName()~`"`, tok.toStringRaw(), state);
					
					// Find and perform goto
					lookupAction(reducedToken);
					if(!actionFound)
						throw new Exception("Couldn't find appropriate goto after reduction.");
					if(chosenAction.type != GoldLALRAction.Type.Goto)
						throw new Exception(
							stformat("Internal Error: After reduction, found action type #{} instead of goto.",
							         chosenAction.type));
					
					state = chosenAction.target;
					//Stdout.formatln(`, {}`, state);
					displayStack();
					break;

				case GoldLALRAction.Type.Goto:
/*					
					state = chosenAction.target;
					Stdout.formatln(`Goto: "{}", state -> {}`,
					                tok.symbolName(), state);
*/
					throw new Exception("Internal Error: Unexpected Goto");

				case GoldLALRAction.Type.Accept:
					if(tokIndex < tokens.length-1)
						throw new Exception(
							stformat("Internal Error: Extra tokens exist beyond end of parsing. Current token: #{} of {}",
							         tokIndex, tokens.length));
					done = true;
					Stdout.formatln(`Accept!`);
					break;

				default:
					throw new Exception(
						stformat("Internal Error: Unknown action type (#{}) found when parsing. LALR state #{}",
								 chosenAction.type, state));
				}
			}
			else
			{
				throw new Exception(
					stformat(`Parse error: LALR state #{}, Token #{} ("{}") "{}"`,
							 state, tokIndex, tok.symbolName(), tok.toStringRaw()));
			}
			
			if(tokenStack.length != stateStack.length)
				throw new Exception(
					stformat("Internal Error: tokenStack[{}] and stateStack[{}] sizes out of sync.",
					         tokenStack.length, stateStack.length));
		}
		
		return tokenStack;
	}
}
